;;; x-face-e21.el --- X-Face utilities for Emacs 21

;; Copyright (C) 2000, 2001, 2002 Katsumi Yamaoka

;; Author: Katsumi Yamaoka  <yamaoka@jpl.org>
;; Created: 2000/06/16
;; Revised: 2002/08/21
;; Keywords: x-face, cmail, emh, gnus, mew, mh-e, rmail, vm, wanderlust, bbdb

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; FYI: Gnus and VM (7.03 and later) can display X-Faces inline under
;; Emacs 21 by themselves.  So, you don't have to use it if you are
;; using one of them and you only want to show X-Faces simply.

;; At first, you have to install this file manually to the location
;; which is referred by `load-path'.  For example:
;;
;; # cp -p x-face-e21.el /usr/local/share/emacs/site-lisp/
;; # cd /usr/local/share/emacs/site-lisp/
;; # emacs -batch -f batch-byte-compile x-face-e21.el

;; If you would like to use BBDB, it should be the version 2.2 or
;; later.  The old versions are out of support even if it works.

;; If you wish to scale down or scale up X-Face images, the external
;; commands pnmscale and pgmtoppm are required.  This feature has been
;; tested with netpbm-9.24 and its bundled commands.

;; Gray X-Faces are now supported.  You can save those X-Face headers
;; in an XPM file using the command `x-face-save'.

;; The command `x-face-insert' allows not only XBM files but also pre-
;; encoded data files.  Therefore, if you already have such a file,
;; for example, ~/.xface which is the default value of
;; `wl-x-face-file' or `mew-x-face-file', you can use it just as it
;; is.

;; Here are some examples of how to use this package with various MUAs
;; and BBDB.  If you are using Emacs versions prior to 19.23 besides
;; Emacs 21, replace the line
;;      ((= 21 emacs-major-version)
;; with
;;      ((and (boundp 'emacs-major-version) (= 21 emacs-major-version))

;;[Common settings for all MUAs] For your .emacs file:
;;----------------------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;; settings for XEmacs.
;;       )
;;      ((= 21 emacs-major-version)
;;       ;;
;;       ;; You may have no need to use the following one line if you
;;       ;; always use T-gnus 6.14.5 revision 07 and later.
;;       (autoload 'x-face-decode-message-header "x-face-e21")
;;       ;;
;;       (autoload 'x-face-insert "x-face-e21" nil t)
;;       (autoload 'x-face-save "x-face-e21" nil t)
;;       (autoload 'x-face-show "x-face-e21" nil t)
;;       (autoload 'x-face-turn-off "x-face-e21")
;;       ;;
;;       ;; Show X-Face images when `x-face-insert' is done.
;;       (setq x-face-auto-image t)
;;       ;;
;;       ;; If you show X-Face images in the message sending buffer,
;;       ;; it is STRONGLY recommended that you remove images from the
;;       ;; buffer before sending a message.  However, it seems not to
;;       ;; be required for Gnusae so far.  The following lines are for
;;       ;; SEMI and Mew.  The latter can be put into .mew file instead.
;;       (add-hook 'mime-edit-translate-hook 'x-face-turn-off)
;;       (add-hook 'mew-make-message-hook 'x-face-turn-off))
;;      (t
;;       ;; settings for the other Emacsen.
;;       ))

;;[Cmail with SEMI] For your .emacs file:
;;---------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((= 21 emacs-major-version)
;;       (add-hook 'mime-display-header-hook 'x-face-decode-message-header)
;;       (eval-after-load "cmail-modes"
;;	 '(progn
;;	    (define-key cmail-summary-mode-map "\C-x4s" 'x-face-save)
;;	    (define-key cmail-mail-mode-map "\C-x4i" 'x-face-insert)
;;	    (define-key cmail-mail-mode-map "\M-t" 'x-face-show)))
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'cmail-mail-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[Gnusae] For your .gnus file:
;;-----------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((= 21 emacs-major-version)
;;       ;;
;;       ;; You may have no need to use the following two lines if you
;;       ;; always use T-gnus 6.14.5 revision 07 and later.
;;       (setq gnus-article-x-face-command 'x-face-decode-message-header
;;	     gnus-treat-display-xface 'head)
;;       ;;
;;       (eval-after-load "gnus-sum"
;;	 '(define-key gnus-summary-mode-map "\C-x4s" 'x-face-save))
;;       (define-key message-mode-map "\C-x4i" 'x-face-insert)
;;       (define-key message-mode-map "\M-t" 'x-face-show)
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'gnus-message-setup-hook 'x-face-insert)
;;       ;;
;;       ;; Uncomment the following line if you would like to show gray
;;       ;; X-Faces using this program instead of the native function.
;;       ;;
;;       ;;(setq gnus-treat-display-grey-xface nil)
;;       )
;;      (t
;;       ;;
;;       ))

;;[Mew] For your .mew file:
;;-------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((= 21 emacs-major-version)
;;       (setq mew-use-highlight-x-face t
;;	     mew-use-highlight-x-face-function 'x-face-decode-message-header)
;;       (define-key mew-summary-mode-map "\C-x4s" 'x-face-save)
;;       (define-key mew-draft-mode-map "\C-x4i" 'x-face-insert)
;;       (define-key mew-draft-header-map "\C-x4i" 'x-face-insert)
;;       (define-key mew-draft-mode-map "\M-t" 'x-face-show)
;;       (define-key mew-draft-header-map "\M-t" 'x-face-show)
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'mew-draft-mode-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[MH-E without EMH] For your .emacs file:
;;----------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((= 21 emacs-major-version)
;;       (add-hook 'mh-show-mode-hook 'x-face-decode-message-header)
;;       (eval-after-load "mh-e"
;;	 '(define-key mh-folder-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "mh-comp"
;;	 '(progn
;;	    (define-key mh-letter-mode-map "\C-x4i" 'x-face-insert)
;;	    (define-key mh-letter-mode-map "\M-t" 'x-face-show)))
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'mh-letter-mode-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[MH-E with EMH] For your .emacs file:
;;-------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((= 21 emacs-major-version)
;;       (add-hook 'mime-display-header-hook 'x-face-decode-message-header)
;;       (eval-after-load "mh-e"
;;	 '(define-key mh-folder-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "mh-comp"
;;	 '(progn
;;	    (define-key mh-letter-mode-map "\C-x4i" 'x-face-insert)
;;	    (define-key mh-letter-mode-map "\M-t" 'x-face-show)))
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'mh-letter-mode-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[RMAIL without RMAIL-MIME] For your .emacs file:
;;------------------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((= 21 emacs-major-version)
;;       (add-hook 'rmail-show-message-hook 'x-face-decode-message-header)
;;       (eval-after-load "rmail"
;;	 '(define-key rmail-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "rmailsum"
;;	 '(define-key rmail-summary-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "sendmail"
;;	 '(progn
;;	    (define-key mail-mode-map "\C-x4i" 'x-face-insert)
;;	    (define-key mail-mode-map "\M-t" 'x-face-show)))
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'mail-setup-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[RMAIL with RMAIL-MIME] For your .emacs file:
;;---------------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((= 21 emacs-major-version)
;;       (add-hook 'mime-display-header-hook 'x-face-decode-message-header)
;;       (eval-after-load "rmail"
;;	 '(define-key rmail-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "rmailsum"
;;	 '(define-key rmail-summary-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "sendmail"
;;	 '(progn
;;	    (define-key mail-mode-map "\C-x4i" 'x-face-insert)
;;	    (define-key mail-mode-map "\M-t" 'x-face-show)))
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'mail-setup-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[VM without SEMI] For your .vm file:
;;------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((= 21 emacs-major-version)
;;       (eval-after-load "vm-page"
;;	 '(defadvice vm-energize-headers-and-xfaces
;;	    (after x-face-decode-message-header activate compile)
;;	    "Show X-Face using `x-face-decode-message-header'."
;;	    (save-restriction
;;	      (widen)
;;	      (narrow-to-region (vm-headers-of (car vm-message-pointer))
;;				(vm-text-of (car vm-message-pointer)))
;;	      (x-face-decode-message-header))))
;;       (define-key vm-summary-mode-map "\C-x4s" 'x-face-save)
;;       (define-key vm-mail-mode-map "\C-x4i" 'x-face-insert)
;;       (define-key vm-mail-mode-map "\M-t" 'x-face-show)
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'vm-mail-mode-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[VM with SEMI] For your .vm file:
;;---------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((= 21 emacs-major-version)
;;       (add-hook 'mime-display-header-hook 'x-face-decode-message-header)
;;       (define-key vm-summary-mode-map "\C-x4s" 'x-face-save)
;;       (define-key vm-mail-mode-map "\C-x4i" 'x-face-insert)
;;       (define-key vm-mail-mode-map "\M-t" 'x-face-show)
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'vm-mail-mode-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[Wanderlust] For your .wl file:
;;-------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((= 21 emacs-major-version)
;;       (setq wl-highlight-x-face-function 'x-face-decode-message-header)
;;       (define-key wl-summary-mode-map "\C-x4s" 'x-face-save)
;;       (define-key wl-draft-mode-map "\C-x4i" 'x-face-insert)
;;       ;; "\M-t" key is reserved for wl command.
;;       (define-key wl-draft-mode-map "\M-\C-t" 'x-face-show)
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'wl-mail-setup-hook 'x-face-insert)
;;       ;; If you use `wl-draft-insert-x-face-field' instead of
;;       ;; `x-face-insert' for inserting an X-Face, you can highlight
;;       ;; it as an image with the setting of the following hook:
;;       (add-hook 'wl-draft-insert-x-face-field-hook
;;		 (lambda nil
;;		   (x-face-insert wl-x-face-file)))
;;       )
;;      (t
;;       ;;
;;       ))

;;[BBDB]
;;------
;; To make BBDB automatically add X-Face headers in the record, put the
;; following lines in your startup (commonly .emacs) file after setting
;; up the value of `bbdb-auto-notes-alist' or after the library
;; "bbdb-hooks" is loaded (you may use `eval-after-load' for that).
;;
;;(put 'face 'field-separator "\n")
;;(setq bbdb-auto-notes-alist
;;      (append bbdb-auto-notes-alist
;;	      `(("x-face"
;;		 (,(concat "[\t\n ]*\\([^\t\n ]*\\)"
;;			   "\\([\t\n ]+\\([^\t\n ]+\\)\\)?"
;;			   "\\([\t\n ]+\\([^\t\n ]+\\)\\)?"
;;			   "\\([\t\n ]+\\([^\t\n ]+\\)\\)?"
;;			   "\\([\t\n ]+\\([^\t\n ]+\\)\\)?")
;;		  face
;;		  "\\1\\3\\5\\7\\9")))))
;;
;; For only Emacs 21, you may use the following instead:
;;
;;(put 'face 'field-separator "\n")
;;(setq bbdb-auto-notes-alist
;;      (append bbdb-auto-notes-alist
;;	      `(("x-face"
;;		 (,(concat "[[:blank:]\n]*\\([^[:blank:]\n]*\\)"
;;			   "\\(?:[[:blank:]\n]+\\([^[:blank:]\n]+\\)\\)?"
;;			   "\\(?:[[:blank:]\n]+\\([^[:blank:]\n]+\\)\\)?"
;;			   "\\(?:[[:blank:]\n]+\\([^[:blank:]\n]+\\)\\)?"
;;			   "\\(?:[[:blank:]\n]+\\([^[:blank:]\n]+\\)\\)?")
;;		  face
;;		  "\\1\\2\\3\\4\\5")))))

;;; Code:

(eval-and-compile
  (if (or (featurep 'xemacs)
	  (not (boundp 'emacs-major-version))
	  (< emacs-major-version 21))
      (error "This program is for FSF Emacs 21+, good-bye")))

(defconst x-face-e21-version "0.118")

(defgroup x-face nil
  "X-Face utilities."
  :group 'news
  :group 'mail)

(defcustom x-face-image-file-directory "~/x-faces/"
  "Name of the directory where the image files can be found."
  :version "21.1"
  :type 'directory
  :group 'x-face)

(defcustom x-face-default-xbm-file nil
  "Default XBM file name for the user's face.  The file will be inserted
in the buffer as an X-Face header using `x-face-insert' without
argument.  If the file name does not include directory component, it
should be found in the directory which is specified by
`x-face-image-file-directory'."
  :version "21.1"
  :type 'string
  :group 'x-face)

(defcustom x-face-insert-query-file-name-when-no-argument nil
  "If it is non-nil, you will be prompted for a file name unconditionally
when `x-face-insert' is called non-interactively without argument."
  :version "21.1"
  :type 'boolean
  :group 'x-face)

(defcustom x-face-image-file-directory-for-save x-face-image-file-directory
  "Name of the directory where image files should go.  It is referred by
the command `x-face-save'."
  :version "21.1"
  :type 'directory
  :group 'x-face)

(defcustom x-face-compressor nil
  "Compressing method used for saving image files.  This variable can be
set to nil, `gzip' or `bzip2'."
  :version "21.1"
  :type '(radio (const :format "%v " bzip2)
		(const :format "%v " gzip)
		(const :tag "off" nil))
  :group 'x-face)

(defcustom x-face-use-overlay nil
  "If it is non-nil, X-Face will be shown as an overlay with
`before-string' property.  It is useful if you don't want to insert
any extra text in the showing buffer.  However, you should take notice
that some MUAs (e.g. Gnus) might delete such overlays unconditionally.
It will be overridden with t buffer-locally when `x-face-insert' is
done."
  :version "21.1"
  :type 'boolean
  :group 'x-face)

(defcustom x-face-hide-related-headers nil
  "If it is non-nil, not only X-Face headers but also all related headers
\(e.g. X-Face-Version) will be hidden in the showing buffer.  It will
be overridden with nil buffer-locally when `x-face-insert' is done."
  :version "21.1"
  :type 'boolean
  :group 'x-face)

(defcustom x-face-hidden-properties '(invisible t intangible t)
  "Property list to hide raw X-Face headers or its related headers.  It
will be modified buffer-locally in order to transfer an icon image
when `x-face-insert' is done."
  :version "21.1"
  :type 'plist
  :group 'x-face)

(defcustom x-face-auto-image nil
  "If it is non-nil, show X-Face images when `x-face-insert' is done.
It can also be a function, for example, `interactive-p',
`(lambda nil (memq major-mode '(message-mode wl-draft-mode)))', etc."
  :version "21.1"
  :type '(radio (const :format "%t " :tag "OFF" nil)
		(const :format "%t\n" :tag "ON" t)
		(function :format "Function: %v"))
  :group 'x-face)

(defcustom x-face-mouse-face 'highlight
  "Face used to highlight buttons when the mouse moves over it."
  :version "21.1"
  :type 'face
  :group 'x-face)

(defcustom x-face-bbdb-display (and (locate-library "bbdb") t)
  "If it is non-nil, show X-Face images in the BBDB buffer."
  :version "21.1"
  :type 'boolean
  :group 'x-face)

(defvar x-face-ring nil "Cache for face data.")
(defvar x-face-ring-size 32 "Length of the ring used for `x-face-ring'.")

(eval-when-compile
  (defmacro x-face-dark-background-p ()
    "Return t if the selected frame's background is dark."
    '(eq 'dark (frame-parameter (selected-frame) 'background-mode))))

(eval-when-compile
  (defmacro x-face-image-attributes-for-bbdb ()
    "Merge the value of the option `x-face-image-attributes-for-bbdb' into
the copy of `x-face-image-attributes' and return the result."
    '(let* ((bgmode (if (x-face-dark-background-p)
			'dark
		      'light))
	    (basis (copy-sequence (symbol-value 'x-face-image-attributes)))
	    (element (assq bgmode basis))
	    (modifiers (cdr (assq bgmode
				  (symbol-value
				   'x-face-image-attributes-for-bbdb))))
	    attributes)
       (if element
	   (progn
	     (setq attributes (copy-sequence (cdr element)))
	     (while modifiers
	       (setq attributes (plist-put attributes
					   (car modifiers) (cadr modifiers))
		     modifiers (cddr modifiers)))
	     (cons (cons bgmode attributes) (delq element basis)))
	 (cons (cons bgmode modifiers) basis)))))

(put 'x-face-internal-function 'show-sample
     (lambda (&optional bbdb remove-cache)
       "Show an X-Face sample image in a customization buffer.  If BBDB is
non-nil, assign the value of `x-face-image-attributes-for-bbdb' to an
image instead of `x-face-image-attributes'.  If REMOVE-CACHE is non-
nil, reset the value of `x-face-ring' to nil."
       (when (and (display-images-p)
		  (image-type-available-p 'pbm)
		  (eq 'custom-mode major-mode))
	 (save-excursion
	   (let ((inhibit-point-motion-hooks t))
	     (goto-char (point-min))
	     (when (re-search-forward "^<?X Face Image Attributes" nil t)
	       (forward-line 1)
	       (let ((inhibit-read-only t))
		 (delete-region (point)
				(progn
				  (when (looking-at "From:")
				    (forward-line 1)
				    (when (search-forward "X-Face:" nil t)
				      (while (progn
					       (forward-line 1)
					       (memq (char-after)
						     '(?\t ?\ ))))))
				  (skip-chars-forward "\n")
				  (point)))
		 (save-restriction
		   (narrow-to-region (point) (point))
		   (unless
		       (and x-face-default-xbm-file
			    (prog2
				(insert "From: "
					(if user-mail-address
					    (if user-full-name
						(concat
						 user-full-name " <"
						 user-mail-address ">")
					      user-mail-address)
					  "X-Face")
					"\n")
				(condition-case nil
				    (let (x-face-auto-image
					  x-face-hidden-properties)
				      (x-face-insert x-face-default-xbm-file)
				      t)
				  (error
				   (delete-region (point-min) (point-max))
				   nil))
			      (dolist (var '(x-face-use-overlay
					     x-face-hide-related-headers
					     x-face-hidden-properties))
				(kill-local-variable var))))
		     (insert "\
From: X-Face
X-Face: 2i'm.M0UyETCme?'R/?fE}i)R-aY$t;].MSLwmUfB\"^3H+so!vO79{mzviSR4#DM+}\"\"
 ZwMOJ~e&Rr*qL'CrQZo-@jdTL=w{o3Pxu2PY]_qB=w%GLU1S_Pk8HX>4C}W2YTc=V=(~QH[vcm2!]O
 pq&CB^,K:NO/lVI-m&kP;pa&K.Xo)<V4=l<eRW]>X\n"))
		   (when remove-cache
		     (setq x-face-ring nil))
		   (if bbdb
		       (let ((x-face-image-attributes
			      (x-face-image-attributes-for-bbdb))
			     x-face-ring)
			 (x-face-decode-message-header))
		     (let (x-face-ring)
		       (x-face-decode-message-header)))
		   (goto-char (point-min))
		   (forward-line 1)
		   (delete-region (point) (point-max))))))))))

(put 'x-face-internal-function 'image-attributes
     (lambda nil
       "Return a widget type for entering X-Face image attributes."
       (let* ((props (list :ascent :margin :relief :conversion
			   :foreground :background
			   ;; Extra keyword.
			   :scale-factor))
	      (lmax (let ((result 0))
		      (dolist (prop props result)
			(setq result (max result
					  (length (symbol-name prop)))))))
	      (tag "When frame background mode is %v")
	      (left (- (length tag) lmax 4))
	      (selection `(checklist :format
				     ,(format (format "%%%ds%%%%v" (- left 2))
					      "")))
	      prop)
	 (while (setq prop (pop props))
	   (setq selection
		 (nconc
		  selection
		  `((list :format "%v" :inline t
			  (const :format "%v " :value ,prop)
			  (sexp :format
				,(format
				  (format
				   "%%%ds%%-%ds"
				   (- lmax (length (symbol-name prop)))
				   (if props left 0))
				  "" "%v")))))))
	 `(list (cons :format "%v" (sexp :format ,tag) ,selection)
		(cons :format "%v" (sexp :format ,tag) ,selection)))))

(eval-when-compile
  (defmacro x-face-cleanup-plist (plist)
    "Remove properties with the value nil in PLIST."
    `(let ((plist ,plist)
	   property value rest)
       (while (prog1
		  (setq property (pop plist))
		(when (setq value (pop plist))
		  (setq rest (nconc rest (list property value))))))
       rest)))

(put 'x-face-internal-function 'cleanup-attributes
     (lambda (attributes)
       "Remove properties with the value nil in each element of ATTRIBUTES."
       (dolist (element attributes attributes)
	 (setcdr element (x-face-cleanup-plist (cdr element))))))

(defcustom x-face-image-attributes
  '((light :ascent 80 :foreground "#000000")
    (dark :ascent 80 :foreground "#000000" :background "#ffffff"))
  "List of image attributes to assign to the X-Face images.  Each element
consists of the value of the frame background mode (`light' or `dark')
and pairs of a keyword and a value.  This allows the extra keyword
`:scale-factor' which is used to scale down or scale up X-Face images.
The value for the keyword `:scale-factor' should be a positive number
or nil (treated as 1)."
  :version "21.1"
  :type (funcall (get 'x-face-internal-function 'image-attributes))
  :get (lambda (symbol)
	 (prog1
	     (let* ((value (default-value symbol))
		    (first (assq (or (frame-parameter (selected-frame)
						      'background-mode)
				     'light)
				 value)))
	       (funcall (get 'x-face-internal-function 'cleanup-attributes)
			(cons first (delq first (copy-sequence value)))))
	   (funcall (get 'x-face-internal-function 'show-sample))))
  :set (lambda (symbol value)
	 (prog1
	     (custom-set-default
	      symbol
	      (funcall (get 'x-face-internal-function 'cleanup-attributes)
		       value))
	   (funcall (get 'x-face-internal-function 'show-sample) nil t)))
  :group 'x-face)

(defcustom x-face-image-attributes-for-bbdb
  '((light :ascent center :foreground "#002000" :background "#fffacd"
	   :relief 2 :scale-factor 0.5)
    (dark :ascent center :foreground "#002000" :background "#fffacd"
	  :relief 2 :scale-factor 0.5))
  "List of attributes to modify the value of `x-face-image-attributes'
which is used to show X-Face images on the BBDB buffer.  The value
form is the same as `x-face-image-attributes'."
  :version "21.1"
  :type (funcall (get 'x-face-internal-function 'image-attributes))
  :get (lambda (symbol)
	 (prog1
	     (let* ((value (default-value symbol))
		    (first (assq (or (frame-parameter (selected-frame)
						      'background-mode)
				     'light)
				 value)))
	       (cons first (delq first (copy-sequence value))))
	   (funcall (get 'x-face-internal-function 'show-sample) t)))
  :set (lambda (symbol value)
	 (prog1
	     (custom-set-default symbol value)
	   (funcall (get 'x-face-internal-function 'show-sample) t t)))
  :group 'x-face)

(defcustom uncompface-program "uncompface"
  "Program used to decode X-Face."
  :version "21.1"
  :type 'string
  :group 'x-face)

(defcustom compface-program "compface"
  "Program used to encode X-Face."
  :version "21.1"
  :type 'string
  :group 'x-face)

(defcustom x-face-pnmscale-program "pnmscale"
  "The pnmscale executable which comes from the netpbm package.  It will
be used to scale down or scale up X-Face images if the value for the
keyword `:scale-factor' in the option `x-face-image-attributes' or
`x-face-image-attributes-for-bbdb' is neither nil nor number 1.  For
quickly scaling, you can alter it to pnmscalefixed."
  :version "21.1"
  :type '(radio (const :format "%v " nil)
		(string))
  :set (lambda (symbol value)
	 (custom-set-default
	  symbol
	  (and (not noninteractive)
	       (stringp value)
	       (with-temp-buffer
		 (set-buffer-multibyte nil)
		 (insert "P1\n1 1\n0\n")
		 (condition-case nil
		     (let ((coding-system-for-read 'binary)
			   (coding-system-for-write 'binary))
		       (call-process-region (point-min) (point-max)
					    value t '(t nil) nil "2")
		       (goto-char (point-min))
		       (let (case-fold-search)
			 (re-search-forward "\
^P5[[:blank:]\n\r]+2[[:blank:]\n\r]+2[[:blank:]\n\r]+" nil t)))
		   (error nil)))
	       value)))
  :group 'x-face)

(defcustom x-face-pgmtoppm-program "pgmtoppm"
  "The pgmtoppm executable which comes from the netpbm package.  It will
be used to colorize the scaled X-Face images."
  :version "21.1"
  :type '(radio (const :format "%v " nil)
		(string))
  :set (lambda (symbol value)
	 (custom-set-default
	  symbol
	  (and (not noninteractive)
	       (stringp value)
	       (with-temp-buffer
		 (set-buffer-multibyte nil)
		 (insert "P1\n2 1\n0 1\n")
		 (condition-case nil
		     (let ((coding-system-for-read 'binary)
			   (coding-system-for-write 'binary))
		       (call-process-region (point-min) (point-max)
					    value t '(t nil) nil
					    "#000000-#ffffff")
		       (goto-char (point-min))
		       (let (case-fold-search)
			 (re-search-forward "\
^P6[[:blank:]\n\r]+2[[:blank:]\n\r]+1[[:blank:]\n\r]+\
\[[:digit:]]+[[:blank:]\n\r]+"
					    nil t)))
		   (error nil)))
	       value)))
  :group 'x-face)

(defvar x-face-field-icon
  (if (image-type-available-p 'xpm)
      (create-image "/* XPM */
static char * x_face_field_icon_xpm[] = {
/* width height ncolors chars_per_pixel */
\"16 14 6 1\",
/* colors */
\"     c None\",
\".    c #000000 s outline\",
\"X    c #ff0000\",
\"o    c #00ff00\",
\"O    c #ffff00\",
\"+    c #00ffff\",
/* pixels */
\"                \",
\"......       ...\",
\".XXXXX.     .o. \",
\" .XXXXX.   .o.  \",
\"  .XXXXX. .o.   \",
\"   .XXXXX.o.    \",
\"    .XXX.O.     \",
\"     .X.OOO.    \",
\"    .+.OOOOO.   \",
\"   .+. .OOOOO.  \",
\"  .+.   .OOOOO. \",
\" .+.     .OOOOO.\",
\"...       ......\",
\"                \"
};
"
		    'xpm t :ascent 'center)
    (create-image "P1
16 14
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 1 1 1 1 1 0 0 0 0 0 0 0 1 0 0
0 0 1 1 1 1 1 0 0 0 0 0 1 0 0 0
0 0 0 1 1 1 1 1 0 0 0 1 0 0 0 0
0 0 0 0 1 1 1 1 1 0 1 0 0 0 0 0
0 0 0 0 0 1 1 1 0 1 0 0 0 0 0 0
0 0 0 0 0 0 1 0 1 1 1 0 0 0 0 0
0 0 0 0 0 1 0 1 1 1 1 1 0 0 0 0
0 0 0 0 1 0 0 0 1 1 1 1 1 0 0 0
0 0 0 1 0 0 0 0 0 1 1 1 1 1 0 0
0 0 1 0 0 0 0 0 0 0 1 1 1 1 1 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
"
		  'pbm t :ascent 'center :mask 'heuristic))
  "Icon image used to iconified raw X-Face header fields.  It can be
`pbm', `xbm' or `xpm' image format, or the value nil not to iconify
the fields.")

(defvar x-face-enable-cache t
  "*Say whether images are cached in `x-face-ring'.")

(autoload 'ring-elements "ring")
(autoload 'ring-insert "ring")

(defvar x-face-show-buffer nil
  "Internal variable used to say where the X-Face images should be
displayed.")
(defvar x-face-working-buffer " *x-face-working*"
  "Temporary buffer for the internal use.")

(defconst x-face-most-positive-fixnum (eval '(lsh -1 -1))
  "Maximum integer for this Emacs.")

(eval-when-compile
  ;; Avoid byte-compile warnings.
  (autoload 'cmail-folder-buffer "cmail-misc")
  (autoload 'cmail-get-page-number-from-summary "cmail-misc")
  (autoload 'cmail-n-page "cmail-misc")
  (autoload 'cmail-show-contents "cmail")
  (autoload 'gnus-summary-select-article "gnus-sum")
  (autoload 'mew-buffer-message "mew")
  (autoload 'mew-summary-display "mew-summary")
  (autoload 'mh-get-msg-num "mh-utils")
  (autoload 'mh-show-msg "mh-utils")
  (autoload 'vm-follow-summary-cursor "vm-motion")
  (autoload 'wl-message-get-original-buffer "wl-message")
  (autoload 'wl-summary-set-message-buffer-or-redisplay "wl-summary")

  (defsubst x-face-buffer-live-p (buffer)
    "Say whether BUFFER is alive or not."
    (and buffer
	 (get-buffer buffer)
	 (buffer-name (get-buffer buffer))))

  (defsubst x-face-narrow-to-head ()
    "Narrow the buffer to the head of the message.  Point is left at the
beginning of the narrowed-to region."
    (narrow-to-region
     (goto-char (point-min))
     (if (let (case-fold-search)
	   (re-search-forward (concat "^$\\|^"
				      (regexp-quote mail-header-separator)
				      "$")
			      nil t))
	 (match-beginning 0)
       (point-max)))
    (goto-char (point-min)))

  (defmacro x-face-set-properties (start end properties &optional mark)
    "If `x-face-use-overlay' is nil, set PROPERTIES and MARK to the text
from START to END, otherwise use overlay instead of text property."
    `(let ((properties ,properties))
       (when properties
	 (let ((start ,start)
	       (end ,end))
	   (if x-face-use-overlay
	       (let ((overlay (make-overlay start end)))
		 (overlay-put overlay 'evaporate t)
		 (overlay-put overlay 'priority x-face-most-positive-fixnum)
		 ,(when mark
		    `(overlay-put overlay ,mark t))
		 (while properties
		   (overlay-put overlay (pop properties) (pop properties))))
	     ,(when mark
		`(put-text-property start end ,mark t))
	     (let ((s1 (1+ start))
		   (e1 (1- end)))
	       (cond ((< s1 e1)
		      (put-text-property start s1 'front-sticky nil)
		      (put-text-property start e1 'rear-nonsticky nil)
		      (put-text-property s1 end 'front-sticky t)
		      (put-text-property e1 end 'rear-nonsticky t))
		     ((= s1 e1)
		      (add-text-properties start s1
					   '(front-sticky nil
							  rear-nonsticky nil))
		      (add-text-properties e1 end
					   '(front-sticky t
							  rear-nonsticky t)))
		     ((= start e1)
		      (add-text-properties start end
					   '(front-sticky nil
							  rear-nonsticky t)))))
	     (add-text-properties start end properties))))))

  (defsubst x-face-remove-x-face-images ()
    "Remove all X-Face images.  The buffer is expected to be narrowed to
just the header of the message."
    (dolist (overlay (overlays-in (point-min) (point-max)))
      (when (overlay-get overlay 'x-face-image)
	(delete-overlay overlay)))
    (let ((inhibit-point-motion-hooks t)
	  (end (point-min))
	  start regions)
      (while (and end
		  (setq start (text-property-any end (point-max)
						 'x-face-image t)))
	(push (cons start
		    (setq end (text-property-not-all start (point-max)
						     'x-face-image t)))
	      regions))
      (let ((mod (buffer-modified-p))
	    (inhibit-read-only t))
	(dolist (region regions)
	  (delete-region (car region) (or (cdr region) (point-max))))
	(set-buffer-modified-p mod))))

  (defsubst x-face-expose-hidden-text ()
    "Expose raw X-Face and its related headers.  The buffer is expected to
be narrowed to just the header of the message."
    (dolist (overlay (overlays-in (point-min) (point-max)))
      (when (overlay-get overlay 'x-face-invisible)
	(delete-overlay overlay)))
    (let ((props (append x-face-hidden-properties '(x-face-invisible t)))
	  (mod (buffer-modified-p))
	  (inhibit-read-only t)
	  (inhibit-point-motion-hooks t)
	  (end (point-min))
	  start)
      (while (and end
		  (setq start (text-property-any end (point-max)
						 'x-face-invisible t)))
	(setq end (text-property-not-all start (point-max)
					 'x-face-invisible t))
	(remove-text-properties start (or end (point-max)) props))
      (set-buffer-modified-p mod)))

  (defmacro x-face-insert-error (error-symbol &rest args)
    "Signal ERROR-SYMBOL with concatenated ARGS."
    (list 'let* (list (list 'args (cons 'list args))
		      '(rest (if (interactive-p)
				 (pop args)
			       "[x-face-insert]")))
	  (list 'signal error-symbol
		'(list (dolist (arg args rest)
			 (setq rest (concat rest " " arg)))))))

  (defmacro x-face-bool-vector-to-string (bool-vector)
    "Convert a 2304-bit bool-vector to a 288-byte binary string."
    `(read (substring (let (print-length print-level)
			(prin1-to-string ,bool-vector))
		      6)))

  (defmacro x-face-string-to-bool-vector (string)
    "Convert a 288-byte binary string to a 2304-bit bool-vector."
    `(read (concat "#&2304" (let (print-length print-level)
			      (prin1-to-string ,string)))))

  (defmacro x-face-cleanup-x-face (x-face &optional with-header)
    "Strip an X-Face header, whitespaces and newlines from X-FACE.  If
WITH-HEADER is non-nil, it assumes X-FACE includes an X-Face header,
otherwise it will examine whether there is an X-Face header."
    (if (or (eq t with-header)
	    (and with-header
		 (eq 'quote (car-safe with-header))))
	`(apply 'concat (cdr (split-string ,x-face)))
      `(let ((x-face (split-string ,x-face)))
	 (if (or ,with-header
		 (string-match "^X-Face\\(-[[:digit:]]+\\)?:$" (car x-face)))
	     (apply 'concat (cdr x-face))
	   (apply 'concat x-face)))))

  (defmacro x-face-cleanup-x-faces (x-faces &optional with-header)
    "Strip an X-Face header, whitespaces and newlines from each X-FACE.  If
WITH-HEADER is non-nil, it assumes each X-FACE includes an X-Face
header, otherwise it will examine whether there is an X-Face header
in each X-FACE."
    (if (or (eq t with-header)
	    (and with-header
		 (eq 'quote (car-safe with-header))))
	`(let ((x-faces ,x-faces)
	       faces)
	   (while x-faces
	     (push (apply 'concat (cdr (split-string (pop x-faces))))
		   faces))
	   (nreverse faces))
      `(let ((x-faces ,x-faces)
	     (header ,with-header)
	     face faces)
	 (while x-faces
	   (if header
	       (push (apply 'concat (cdr (split-string (car x-faces)))) faces)
	     (if (string-match "^X-Face\\(-[[:digit:]]+\\)?:$"
			       (car (setq face (split-string (car x-faces)))))
		 (progn
		   (setq header t)
		   (push (apply 'concat (cdr face)) faces))
	       (push (apply 'concat face) faces)))
	   (setq x-faces (cdr x-faces)))
	 (nreverse faces)))))

(defconst x-face-mirror (eval-when-compile
			  (let ((table (make-vector 256 nil))
				(i 0)
				j k l)
			    (while (< i 256)
			      (setq j i
				    k 0
				    l 8)
			      (while (> l 0)
				(setq k (lsh k -1)
				      l (1- l))
				(if (>= (setq j (lsh j 1)) 256)
				    (setq j (logand 255 j)
					  k (logior 128 k))))
			      (aset table i k)
			      (setq i (1+ i)))
			    table))
  "Bitwise symmetrical table for 8-bit data.")

(defconst x-face-xpm-colors
  (eval-when-compile
    (vconcat
     " .XoO+@#$%&*=-;:>,<1234567890qwertyuipasdfghjklzxcvbnmMNBVCZASDF"))
  "Color map used to 6-bit XPM images.")

(defun x-face-possibly-change-buffer ()
  "Possibly change the buffer to where the raw X-Face headers can be
found.  You can redefine it for customization.  The value of the
variable `x-face-show-buffer' can be modified if it is needed."
  (cond ((and (eq 'rmail-summary-mode major-mode)
	      (not (featurep 'rmail-mime)))
	 (set-buffer (symbol-value 'rmail-buffer))
	 (x-face-possibly-change-buffer))
	((and (eq 'rmail-mode major-mode)
	      (not (featurep 'rmail-mime)))
	 (let ((buffer (get-buffer-create x-face-working-buffer)))
	   (save-excursion
	     (goto-char (point-min))
	     (save-restriction
	       (widen)
	       (when (search-backward "\n\n" nil t)
		 (skip-chars-backward "\n")
		 (forward-char))
	       (copy-to-buffer buffer (point)
			       (if (search-backward "\n\n" nil t)
				   (match-end 0)
				 (point-min)))))
	   (set-buffer buffer)))))

(defun x-face-possibly-change-buffer-for-saving ()
  "Possibly change the buffer to where the raw X-Face headers can be
found.  This function will be used for saving X-Face headers.  You can
redefine it for customization."
  (cond ((eq 'cmail-summary-mode major-mode)
	 (let ((page (cmail-get-page-number-from-summary))
	       (buffer (get-buffer-create x-face-working-buffer)))
	   (unless (eq page (symbol-value '*cmail-current-page))
	     (cmail-show-contents page))
	   (set-buffer (cmail-folder-buffer
			(symbol-value 'cmail-current-folder)))
	   (cmail-n-page page)
	   (copy-to-buffer buffer (point) (if (search-forward "\n\n" nil t)
					      (1+ (match-beginning 0))
					    (point-max)))
	   (set-buffer buffer)))
	((eq 'gnus-summary-mode major-mode)
	 (gnus-summary-select-article)
	 (set-buffer (symbol-value 'gnus-original-article-buffer)))
	((and (boundp 'gnus-article-buffer)
	      (eq (current-buffer)
		  (get-buffer (symbol-value 'gnus-article-buffer)))
	      (x-face-buffer-live-p (symbol-value
				     'gnus-original-article-buffer)))
	 (set-buffer (symbol-value 'gnus-original-article-buffer)))
	((memq major-mode '(mew-summary-mode mew-virtual-mode))
	 (mew-summary-display nil)
	 (set-buffer (mew-buffer-message)))
	((eq 'mh-folder-mode major-mode)
	 (let* ((num (mh-get-msg-num t))
		(buffer (symbol-value 'mh-show-buffer))
		(folder (symbol-value 'mh-current-folder))
		(raw-buffer (concat "article-" (if (bufferp folder)
						   (buffer-name folder)
						 folder))))
	   (if (and (x-face-buffer-live-p buffer)
		    (string-match (format "/%d$" num)
				  (buffer-file-name (get-buffer buffer))))
	       (set-buffer (if (x-face-buffer-live-p raw-buffer)
			       raw-buffer
			     buffer))
	     (mh-show-msg num)
	     (if (x-face-buffer-live-p raw-buffer)
		 (set-buffer raw-buffer)
	       (current-buffer)))))
	((eq 'rmail-summary-mode major-mode)
	 (set-buffer (symbol-value 'rmail-buffer))
	 (x-face-possibly-change-buffer-for-saving))
	((eq 'rmail-mode major-mode)
	 (let ((buffer (get-buffer-create x-face-working-buffer)))
	   (save-excursion
	     (goto-char (point-min))
	     (save-restriction
	       (widen)
	       (when (search-backward "\n\n" nil t)
		 (skip-chars-backward "\n")
		 (forward-char))
	       (copy-to-buffer buffer (point)
			       (if (search-backward "\n\n" nil t)
				   (match-end 0)
				 (point-min)))))
	   (set-buffer buffer)))
	((eq 'vm-summary-mode major-mode)
	 (vm-follow-summary-cursor)
	 (if (x-face-buffer-live-p (symbol-value 'vm-mail-buffer))
	     (set-buffer (symbol-value 'vm-mail-buffer))
	   (error "Folder buffer has been killed")))
	((eq 'wl-summary-mode major-mode)
	 (wl-summary-set-message-buffer-or-redisplay)
	 (set-buffer (wl-message-get-original-buffer)))))

(defun x-face-sort-gray-x-faces (x-faces)
  "Sort gray X-FACES in order from msb to lsb.  Each X-FACE should have a
header X-Face or X-Face-NUMBER."
  (let (face faces)
    (while x-faces
      (setq face (split-string (pop x-faces)))
      (push (cons (string-to-number (substring (car face) 6 -1))
		  (apply 'concat (cdr face)))
	    faces))
    (setq faces (sort faces 'car-less-than-car))
    (while faces
      (push (cdr (pop faces)) x-faces)))
  x-faces)

;;;###autoload
(defun x-face-to-bitmap (x-face &optional bool-vector bit-reverse cleaned)
  "Convert an X-FACE to raw bitmap data.  If BOOL-VECTOR is non-nil, it
will return a 2304-bit bool-vector, otherwise 288-byte binary data as
a string.  If BIT-REVERSE is non-nil, each byte will be bit-reversed.
If CLEANED is non-nil, X-FACE includes neither an X-Face header,
whitespaces nor newlines."
  (unless cleaned
    (if (stringp x-face)
	(setq x-face (x-face-cleanup-x-face x-face))
      ;; Return an empty X-Face by default.
      (setq x-face ",\\m{?h\\)X")))
  (with-temp-buffer
    (insert x-face)
    (call-process-region (point-min) (point-max)
			 uncompface-program t '(t nil))
    (goto-char (point-min))
    (setq case-fold-search t)
    (while (search-forward "0x" nil t)
      (replace-match ""))
    (goto-char (point-min))
    (while (re-search-forward "[^[:xdigit:]]+" nil t)
      (replace-match ""))
    (goto-char (point-min))
    (insert "\"")
    (while (not (eobp))
      (insert "\\x")
      (forward-char 2))
    (insert "\"")
    (goto-char (point-min))
    (if bit-reverse
	(let ((data (read (current-buffer))))
	  (dolist (byte (prog1
			    (append data nil)
			  (setq data nil)))
	    (push (aref x-face-mirror byte) data))
	  (if bool-vector
	      (x-face-string-to-bool-vector (concat (nreverse data)))
	    (concat (nreverse data))))
      (if bool-vector
	  (x-face-string-to-bool-vector (read (buffer-string)))
	(string-as-unibyte (read (current-buffer)))))))

;;;###autoload
(defun x-face-bitmap-to-pbm (bitmap &optional plain)
  "Convert a raw BITMAP to a PBM format.  BITMAP may be a 2304-bit
bool-vector or a 288-byte binary string.  If PLAIN is non-nil, it will
return a plain PBM format, otherwise a raw PBM format."
  (if plain
      (let ((ix 0)
	    iy
	    (iz 6)
	    byte)
	(when (bool-vector-p bitmap)
	  (setq bitmap (x-face-bool-vector-to-string bitmap)))
	(with-temp-buffer
	  (insert "P1\n48 48\n")
	  (while (< ix 288)
	    (setq byte (aref bitmap ix)
		  ix (1+ ix)
		  iy 128)
	    (while (> iy 0)
	      (insert (if (zerop (logand byte iy))
			  "0"
			"1"))
	      (setq iy (lsh iy -1)))
	    (when (zerop (setq iz (1- iz)))
	      (insert "\n")
	      (setq iz 6)))
	  (buffer-string)))
    (concat "P4\n48 48\n" (if (bool-vector-p bitmap)
			      (x-face-bool-vector-to-string bitmap)
			    bitmap))))

;;;###autoload
(defun x-face-to-pbm (x-face &optional plain cleaned)
  "Convert an X-FACE to a PBM format.  If PLAIN is non-nil, it will
return a plain PBM format, otherwise a raw PBM format.  If CLEANED is
non-nil, X-FACE includes neither an X-Face header, whitespaces nor
newlines."
  (x-face-bitmap-to-pbm (x-face-to-bitmap x-face nil nil cleaned) plain))

;;;###autoload
(defun x-face-bitmap-to-xbm (bitmap &optional filename)
  "Convert a raw BITMAP to an XBM format.  BITMAP may be a 2304-bit
bool-vector or a 288-byte binary string.  FILENAME is used to be
identifiers in a file which defaults to X_Face."
  (when (bool-vector-p bitmap)
    (setq bitmap (x-face-bool-vector-to-string bitmap)))
  (setq filename (if filename
		     (replace-regexp-in-string "[\000-/:-\?[-^`{-\377]" "_"
					       filename)
		   "X_Face"))
  (with-temp-buffer
    (insert "\
#define " filename "_width 48
#define " filename "_height 48
static char " filename "_bits[] = {
 ")
    (let ((ix 0)
	  (iy 15))
      (while (< ix 288)
	(insert (format "0x%02x," (aref x-face-mirror (aref bitmap ix))))
	(setq ix (1+ ix))
	(when (zerop (setq iy (1- iy)))
	  (insert "\n ")
	  (setq iy 15)))
      (delete-backward-char 1)
      (insert "};\n"))
    (buffer-string)))

;;;###autoload
(defun x-face-to-xbm (x-face &optional filename cleaned)
  "Convert an X-FACE to an XBM format.  FILENAME is used to be
identifiers in a file which defaults to X_Face.  If CLEANED is non-nil,
X-FACE includes neither an X-Face header, whitespaces nor newlines."
  (x-face-bitmap-to-xbm (x-face-to-bitmap x-face nil nil cleaned) filename))

;;;###autoload
(defun x-face-gray-x-faces-to-pixmap (x-faces &optional sorted)
  "Convert gray X-FACES to a pixmap.  Pixmap is a vector which contains
288-pixel raw gray map.  Each X-FACE is an encoded X-Face string which
may or may not include an X-Face header, whitespaces or newlines.  If
SORTED is non-nil, it assumes that X-FACES are sorted in order from
msb to lsb."
  (setq x-faces (if sorted
		    (x-face-cleanup-x-faces x-faces)
		  (x-face-sort-gray-x-faces x-faces)))
  (dolist (x-face (prog1
		      x-faces
		    (setq x-faces nil)))
    (push (x-face-to-bitmap x-face t t t) x-faces))
  (setq x-faces (nreverse x-faces))
  (let* ((ix 0)
	 iy
	 (depth (length x-faces))
	 (pixmap (make-vector 2304 nil))
	 pixel)
    (while (< ix 2304)
      (setq iy 0
	    pixel 0)
      (while (< iy depth)
	(setq pixel (if (aref (nth iy x-faces) ix)
			(+ pixel pixel 1)
		      (* pixel 2))
	      iy (1+ iy)))
      (aset pixmap ix pixel)
      (setq ix (1+ ix)))
    pixmap))

;;;###autoload
(defun x-face-gray-pixmap-to-pgm (pixmap &optional maxval plain)
  "Convert a gray PIXMAP to a PGM format.  PIXMAP is a vector which
should contain 288-byte raw gray map.  MAXVAL is a number which
specifies the white level.  If MAXVAL is nil, it will be guessed by
the values of PIXMAP.  If PLAIN is non-nil, it will return a plain PGM
format, otherwise a raw PGM format."
  (unless maxval
    (setq maxval (1- (expt 2 (1+ (logb (apply 'max (append pixmap nil))))))))
  (if plain
      (concat "P2\n48 48\n" (number-to-string maxval) "\n"
	      (mapconcat 'number-to-string pixmap " ") "\n")
    (concat "P5\n48 48\n" (number-to-string maxval) "\n" pixmap)))

;;;###autoload
(defun x-face-gray-pixmap-to-xpm (pixmap &optional filename ncolors)
  "Convert a gray PIXMAP to an XPM format.  PIXMAP is a vector which
should contain 288-byte raw gray map.  FILENAME is used to be
identifiers in a file which defaults to X_Face.  NCOLORS is a number
which specifies how many colors PIXMAP has.  If NCOLORS is nil, it
will be guessed by PIXMAP itself.  Note that it limits the image depth
maximum to 6-bit, thus the lower bits will be ignored."
  (unless ncolors
    (setq ncolors (expt 2 (1+ (logb (apply 'max (append pixmap nil)))))))
  (let ((ix 0)
	iy scale coeff cval)
    (if (> ncolors 64)
	(setq scale (/ 64.0 ncolors)
	      ncolors 64
	      coeff (/ 256.0 63))
      (setq coeff (/ 256.0 (1- ncolors))))
    (with-temp-buffer
      (insert "\
/* XPM */
static char * "
	      (if filename
		  (replace-regexp-in-string "[\000-/:-\?[-^`{-\377]" "_"
					    filename)
		"X_Face")
	      "[] = {
/* width height ncolors chars_per_pixel */
\"48 48 " (number-to-string ncolors) " 1\",
/* colors */
")
      (while (< ix ncolors)
	(setq cval (min (truncate (* ix coeff)) 255))
	(insert (format "\"%c c #%02x%02x%02x\",\n"
			(aref x-face-xpm-colors (- ncolors ix 1))
			cval cval cval))
	(setq ix (1+ ix)))
      (insert "\
/* pixels */
\"")
      (setq ix 0
	    iy 48)
      (if scale
	  (while (< ix 2304)
	    (insert (aref x-face-xpm-colors
			  (- ncolors (truncate (* (aref pixmap ix) scale)) 1)))
	    (setq ix (1+ ix))
	    (when (zerop (setq iy (1- iy)))
	      (insert "\",\n\"")
	      (setq iy 48)))
	(while (< ix 2304)
	  (insert (aref x-face-xpm-colors (- ncolors (aref pixmap ix) 1)))
	  (setq ix (1+ ix))
	  (when (zerop (setq iy (1- iy)))
	    (insert "\",\n\"")
	    (setq iy 48))))
      (delete-backward-char 3)
      (insert "\n};\n")
      (buffer-string))))

;;;###autoload
(defun x-face-gray-x-faces-to-xpm (x-faces &optional sorted filename)
  "Convert gray X-FACES to an XPM format.  Each X-FACE is an encoded
X-Face string which may or may not include an X-Face header,
whitespaces or newlines.  If SORTED is non-nil, it assumes that
X-FACES are sorted in order from msb to lsb.  FILENAME is used to be
identifiers in a file which defaults to X_Face."
  (x-face-gray-pixmap-to-xpm
   (x-face-gray-x-faces-to-pixmap x-faces sorted)
   filename (expt 2 (length x-faces))))

;;;###autoload
(defun x-face-create-image (x-face &rest props)
  "Create a PBM image from X-FACE.  X-FACE is an encoded X-Face string
which may or may not include an X-Face header, whitespaces or newlines.
The rest PROPS are additional image attributes to assign to the image.
The value of the variable `x-face-image-attributes' will be used to
the default image attributes.

Here are some examples of how to use this function:

;; Insert an image at point.
\(insert-image (x-face-create-image X-Face :KEYWORD VALUE ...))

;; Create an image without any properties.
\(let (x-face-image-attributes)
  (x-face-create-image X-Face))

;; Create a scaled image.
\(x-face-create-image X-Face :scale-factor 0.707)

;; Insert an image as an overlay.
\(let ((overlay (make-overlay (point) (1+ (point))))
      (image (x-face-create-image X-Face)))
  (overlay-put overlay 'evaporate t)
  (overlay-put overlay 'before-string (propertize \" \" 'display image)))
"
  (if (stringp x-face)
      (setq x-face (x-face-cleanup-x-face x-face))
    ;; Return an empty X-Face by default.
    (setq x-face ",\\m{?h\\)X"))
  (let* ((params (copy-sequence (cdr (assq (if (x-face-dark-background-p)
					       'dark
					     'light)
					   x-face-image-attributes))))
	 (scale (or (plist-get props :scale-factor)
		    (plist-get params :scale-factor)
		    1))
	 bg fg should-colorize siblings image)
    (while props
      (setq params (plist-put params (pop props) (pop props))))
    (setq params (x-face-cleanup-plist (plist-put params :scale-factor nil))
	  bg (or (plist-get params :background)
		 (face-background 'default)
		 "#ffffff")
	  fg (or (plist-get params :foreground)
		 (face-foreground 'default)
		 "#000000")
	  should-colorize
	  (and x-face-pgmtoppm-program
	       (not (and (string-match
			  "^white$\\|^gr[ae]y100$\\|^#\\(?:fff\\)+$" bg)
			 (string-match
			  "^black$\\|^gr[ae]y0$\\|^#\\(?:000\\)+$" fg)))))
    (plist-put params :background bg)
    (plist-put params :foreground fg)
    (when x-face-enable-cache
      (if x-face-ring
	  (setq siblings (cdr (assoc x-face (ring-elements x-face-ring)))
		image (cdr (assoc scale siblings)))
	(setq x-face-ring (make-ring x-face-ring-size))))
    (unless image
      ;; Create an image.
      (setq image
	    (if (and x-face-pnmscale-program
		     (/= 1 scale))
		(progn
		  (setq image (x-face-to-bitmap x-face nil nil t))
		  (let ((coding-system-for-read 'binary)
			(coding-system-for-write 'binary)
			default-enable-multibyte-characters)
		    (with-temp-buffer
		      (insert "P4\n48 48\n" image)
		      (call-process-region (point-min) (point-max)
					   x-face-pnmscale-program
					   t '(t nil) nil
					   (number-to-string scale))
		      (when should-colorize
			(call-process-region (point-min) (point-max)
					     x-face-pgmtoppm-program
					     t '(t nil) nil
					     (concat fg "-" bg)))
		      (create-image (buffer-string) 'pbm t))))
	      (create-image (x-face-to-pbm x-face) 'pbm t)))
      (when x-face-enable-cache
	(if siblings
	    (push (cons scale image) siblings)
	  (ring-insert x-face-ring (list x-face (cons scale image))))))
    (append image params)))

;;;###autoload
(defun x-face-create-gray-image (x-faces &optional sorted)
  "Convert gray X-FACES to an image in the PGM format.  Each X-FACE
should be cleaned up that an X-Face header has been stripped.  For
caching images, stripping also whitespaces and newlines from each
X-FACE is recommended.  If SORTED is non-nil, it assumes that X-FACES
are sorted in order from msb to lsb."
  (setq x-faces (if sorted
		    (x-face-cleanup-x-faces x-faces)
		  (x-face-sort-gray-x-faces x-faces)))
  (let* ((params (copy-sequence (cdr (assq (if (x-face-dark-background-p)
					       'dark
					     'light)
					   x-face-image-attributes))))
	 (size (or (plist-get params :scale-factor)
		   1))
	 siblings image)
    (dolist (face (prog1
		      (nreverse x-faces)
		    (setq x-faces nil)))
      (push (apply 'concat (split-string face)) x-faces))
    (when x-face-enable-cache
      (if x-face-ring
	  (setq siblings (cdr (assoc x-faces (ring-elements x-face-ring)))
		image (cdr (assoc size siblings)))
	(setq x-face-ring (make-ring x-face-ring-size))))
    (unless image
      (if (and x-face-pnmscale-program
	       (/= 1 size))
	  (let ((coding-system-for-read 'binary)
		(coding-system-for-write 'binary)
		default-enable-multibyte-characters)
	    (with-temp-buffer
	      (insert (x-face-gray-pixmap-to-pgm
		       (x-face-gray-x-faces-to-pixmap x-faces t)
		       (1- (expt 2 (length x-faces)))))
	      (call-process-region (point-min) (point-max)
				   x-face-pnmscale-program t '(t nil) nil
				   (number-to-string size))
	      (setq image (create-image (buffer-string) 'pbm t))))
	(setq image (create-image (x-face-gray-pixmap-to-pgm
				   (x-face-gray-x-faces-to-pixmap x-faces t)
				   (1- (expt 2 (length x-faces))))
				  'pbm t)))
      (when x-face-enable-cache
	(if siblings
	    (push (cons size image) siblings)
	  (ring-insert x-face-ring (list x-faces (cons size image))))))
    (plist-put params :background nil)
    (plist-put params :foreground nil)
    (plist-put params :scale-factor nil)
    (append image (x-face-cleanup-plist params))))

(eval-when-compile
  (defvar last))

;;;###autoload
(defun x-face-decode-message-header (&optional beg end buffer)
  "Display X-Faces in the current message.  Optional BEG and END are
ignored.  If optional BUFFER is specified, it is assumed that the raw
X-Face headers can be found in the BUFFER.  BUFFER can also be a
function which is similar to `x-face-possibly-change-buffer'.  This
requires support for images in your Emacs and the external program
`uncompface'.  (Optional BEG and END are no more than placeholders to
ensure the compatibility with the old programs.)"
  (when (and (display-images-p)
	     (image-type-available-p 'pbm))
    (let ((inhibit-point-motion-hooks t)
	  start gray images)
      (save-excursion
	;; Remove all existing X-Face images.
	(save-restriction
	  (x-face-narrow-to-head)
	  (x-face-remove-x-face-images))
	(setq x-face-show-buffer (current-buffer))
	;; Change buffer to where the raw X-Face headers can be found.
	(cond ((functionp buffer)
	       (funcall buffer))
	      ((x-face-buffer-live-p buffer)
	       (set-buffer buffer))
	      (t
	       (x-face-possibly-change-buffer)))
	;; Extract X-Faces in header.
	(save-restriction
	  (x-face-narrow-to-head)
	  (let ((case-fold-search t)
		x-face x-faces)
	    (while (progn
		     (while (not (or (eobp)
				     (looking-at
				      "X-Face\\(-[[:digit:]]+\\)?:")))
		       (forward-line 1))
		     (not (eobp)))
	      (setq start (point))
	      (when (match-beginning 1)
		(setq gray t))
	      (while (progn
		       (forward-line 1)
		       (memq (char-after) '(?\t ?\ ))))
	      (setq x-face (buffer-substring-no-properties start (point)))
	      (store-match-data nil)
	      (string-match "\\([\000- ]\\)+\\([!-~]\\)?" x-face)
	      ;; If a header has an empty body, use the encoded empty data.
	      (push (if (match-beginning 2)
			x-face
		      (concat (substring x-face 0 (match-beginning 0))
			      " ,\\m{?h\\)X"))
		    x-faces))
	    (if gray
		(setq images (list (x-face-create-gray-image x-faces)))
	      (while x-faces
		(push (x-face-create-image
		       (apply 'concat (cdr (split-string (pop x-faces)))))
		      images))))))
      (when (x-face-buffer-live-p x-face-working-buffer)
	(kill-buffer x-face-working-buffer))
      (when images
	;; Insert X-Face images.
	(save-excursion
	  (set-buffer x-face-show-buffer)
	  (let ((case-fold-search t)
		(inhibit-read-only t)
		(mod (buffer-modified-p)))
	    (save-restriction
	      (x-face-narrow-to-head)
	      (let ((from (if (re-search-forward "^From:" nil t)
			      (prog1
				  (match-end 0)
				(goto-char (point-min)))
			    (point))))
		;; Hide raw X-Faca headers.
		(let ((icon (plist-get x-face-hidden-properties 'display)))
		  (when icon
		    (setq icon (copy-sequence icon))
		    (let ((type (plist-get (cdr icon) :type))
			  (dark (x-face-dark-background-p))
			  (bg (plist-get (cdar images) :background))
			  (fg (plist-get (cdar images) :foreground)))
		      (cond ((eq 'xpm type)
			     (plist-put (cdr icon) :color-symbols
					(list (cons "outline"
						    (if dark
							bg
						      fg)))))
			    ((memq type '(pbm xbm))
			     (plist-put (cdr icon) :foreground
					(if dark
					    bg
					  fg))))))
		  (let ((regexp (if x-face-hide-related-headers
				    "X-Face.*: ?"
				  "X-Face\\(-[[:digit:]]+\\)?: ?")))
		    (while (progn
			     (while (not (or (eobp) (looking-at regexp)))
			       (forward-line 1))
			     (not (eobp)))
		      (setq start (if icon
				      (match-end 0)
				    (point)))
		      (while (progn
			       (forward-line 1)
			       (memq (char-after) '(?\t ?\ ))))
		      (if icon
			  (x-face-set-properties start (1- (point))
						 (plist-put
						  (copy-sequence
						   x-face-hidden-properties)
						  'display icon)
						 'x-face-invisible)
			(if (and (boundp 'gnus-article-buffer)
				 (eq (current-buffer)
				     (get-buffer (symbol-value
						  'gnus-article-buffer))))
			    ;; We should not put the hidden properties at
			    ;; the last newline since it will be left when
			    ;; sorting header is done.
			    (x-face-set-properties start (1- (point))
						   x-face-hidden-properties
						   'x-face-invisible)
			  (x-face-set-properties start (point)
						 x-face-hidden-properties
						 'x-face-invisible))))))
		(goto-char from)))
	    (if x-face-use-overlay
		(let ((overlay (make-overlay (point) (1+ (point))))
		      (rest ""))
		  (setq images (nreverse images))
		  (while images
		    ;; There is a face `default' used to hide an
		    ;; underline text behind an image.
		    (setq rest (concat (propertize " " 'display
						   (pop images) 'face 'default)
				       rest)))
		  (overlay-put overlay 'before-string
			       (if (bolp)
				   (concat rest "\n")
				 rest))
		  (overlay-put overlay 'x-face-image t)
		  (overlay-put overlay 'priority x-face-most-positive-fixnum)
		  (overlay-put overlay 'evaporate t))
	      (while images
		(setq start (point))
		(when (prog1
			  (when (bolp)
			    (insert "From:")
			    t)
			(insert-image (pop images)))
		  (insert " X-Face\n"))
		(put-text-property start (point) 'x-face-image t)))
	    (set-buffer-modified-p mod)))))
    (when (and (boundp 'gnus-article-buffer)
	       (eq (current-buffer) (symbol-value 'gnus-article-buffer)))
      (set 'last t))))

;;;###autoload
(defun x-face-show (&optional arg)
  "Toggle showing X-Face images.  With ARG, turn showing on if and only
if ARG is positive."
  (interactive "P")
  (if (if (numberp arg)
	  (> arg 0)
	(not (or (and (numberp arg) (< arg 0))
		 ;; There is no From header in the buffer.
		 (get-text-property (point-min) 'x-face-image)
		 (save-excursion
		   (save-restriction
		     (x-face-narrow-to-head)
		     (or (let ((overlays (overlays-in (point-min)
						      (point-max))))
			   (while (and overlays
				       (not (overlay-get (car overlays)
							 'x-face-image)))
			     (setq overlays (cdr overlays)))
			   overlays)
			 (text-property-any (point-min) (point-max)
					    'x-face-image t)))))))
      ;; The non-nil value for `message-strip-special-text-properties'
      ;; (by default) prevents hiding raw X-Face headers (some users
      ;; likely show her/his own face in the message buffer :-).
      (let (message-strip-special-text-properties)
	(x-face-decode-message-header))
    (x-face-remove-x-face-images)
    (x-face-expose-hidden-text)))

;;;###autoload
(defun x-face-turn-off ()
  "Remove X-face images from the buffer."
  (x-face-show -1))

(defvar x-face-read-file-name-type 'xbm
  "Internal variable used to say what a type of image files should be
read by the function `x-face-read-file-name'.")

(defun x-face-read-file-name-completion-handler (operation &rest args)
  "Internal function used to handle completion for image files.  The
variable `x-face-read-file-name-type' controls what a type of image
files should be handled."
  (let ((inhibit-file-name-handlers
	 (cons 'x-face-read-file-name-completion-handler
	       inhibit-file-name-handlers))
	(inhibit-file-name-operation operation)
	(partial (car args))
	(dir (cadr args))
	(regexp (format "\\(?:/\\|\\.%s\\(?:\\.bz2\\|\\.gz\\)?\\)\\'"
			x-face-read-file-name-type))
	(case-fold-search t)
	candidates)
    (if (eq 'file-name-completion operation)
	(try-completion
	 partial
	 (dolist (name (file-name-all-completions partial dir) candidates)
	   (setq candidates (nconc candidates (list (list name))))))
      (if (eq 'file-name-all-completions operation)
	  (dolist (name (file-name-all-completions partial dir) candidates)
	    (when (string-match regexp name)
	      (setq candidates (nconc candidates (list name)))))
	(apply operation args)))))

;;;###autoload
(defun x-face-read-file-name (prompt &optional dir default-filename
				     mustmatch initial)
  "Read an image file name.  This function is equivalent to
`read-file-name', except that it limits to image files.  The variable
`x-face-read-file-name-type' controls what a type of an image file
should be read.  Note that since the variable `file-name-history' will
not be updated in this function, use `x-face-put-file-name-in-history'
to register the actual file name after it has been decided."
  (let ((file-name-handler-alist
	 (cons (cons "" 'x-face-read-file-name-completion-handler)
	       file-name-handler-alist))
	(regexp (format "\\(?:/\\|\\.%s\\(?:\\.bz2\\|\\.gz\\)?\\)\\'"
			x-face-read-file-name-type))
	(case-fold-search t)
	(file-name-history file-name-history)
	file history)
    (while (setq file (pop file-name-history))
      (when (string-match regexp file)
	(push file history)))
    (setq file-name-history (nreverse history))
    (read-file-name prompt
		    ;; The 2nd arg DIR should have the trailing "/"
		    ;; because the 5th arg INITIAL will be appended it.
		    (file-name-as-directory dir)
		    default-filename mustmatch initial)))

(defun x-face-put-file-name-in-history (filename)
  "Put FILENAME in `file-name-history'."
  (let ((abbrev (when (and abbreviated-home-dir
			   (string-match abbreviated-home-dir filename))
		  (replace-match "~/" nil nil filename))))
    (unless (or (member filename file-name-history)
		(and abbrev
		     (member abbrev file-name-history)))
      (push (or abbrev filename) file-name-history))))

;;;###autoload
(defun x-face-insert (&optional xbm-file keep-existing-headers)
  "Insert XBM-FILE as an X-Face header.  If XBM-FILE is omitted, the
value of `x-face-default-xbm-file' will be used for it.  If the
contents of a file do not look like the XBM format, they will be
regarded as pre-encoded data.  If optional KEEP-EXISTING-HEADERS is
non-nil, existing X-Face headers will not be removed.  This requires
the external program `compface'.  It will override the value of the
variable `x-face-use-overlay' to t buffer-locally."
  (interactive "*i\nP")
  (unless (and (stringp xbm-file) (> (length xbm-file) 0))
    (let (dir default)
      (when (and (stringp x-face-default-xbm-file)
		 (> (length x-face-default-xbm-file) 0))
	(setq dir (file-name-directory x-face-default-xbm-file)
	      default
	      (if (and dir
		       (let ((inode1 (nth 10 (file-attributes dir)))
			     (inode2 (nth 10 (file-attributes
					      x-face-image-file-directory))))
			 (and inode1 inode2 (equal inode1 inode2))))
		  (file-name-nondirectory x-face-default-xbm-file)
		x-face-default-xbm-file)))
      (if (or (interactive-p)
	      x-face-insert-query-file-name-when-no-argument)
	  (setq xbm-file
		(let ((x-face-read-file-name-type 'xbm))
		  (x-face-read-file-name
		   (concat "XBM file"
			   (if default
			       (concat " (default " default "): ")
			     ": "))
		   (or dir x-face-image-file-directory) default t)))
	(when default
	  (setq xbm-file x-face-default-xbm-file))))
    (unless (and (stringp xbm-file) (> (length xbm-file) 0))
      (x-face-insert-error 'file-error "No xbm file is specified")))
  (setq xbm-file (expand-file-name xbm-file x-face-image-file-directory))
  (when (interactive-p)
    (x-face-put-file-name-in-history xbm-file))
  (when (file-directory-p xbm-file)
    (x-face-insert-error 'file-error xbm-file "is a directory"))
  (or (file-exists-p xbm-file)
      (and (file-exists-p (concat xbm-file ".bz2"))
	   (setq xbm-file (concat xbm-file ".bz2")))
      (and (file-exists-p (concat xbm-file ".gz"))
	   (setq xbm-file (concat xbm-file ".gz")))
      (x-face-insert-error 'file-error "Cannot open XBM file:" xbm-file))
  (let ((x-face
	 (with-temp-buffer
	   (set-buffer-multibyte nil)
	   (if (and (string-match "\\.bz2$\\|\\.gz$" xbm-file)
		    (not auto-compression-mode))
	       (progn
		 (auto-compression-mode 1)
		 (unwind-protect
		     (insert-file-contents xbm-file)
		   (auto-compression-mode -1)))
	     (if auto-image-file-mode
		 (progn
		   (auto-image-file-mode -1)
		   (insert-file-contents xbm-file)
		   (auto-image-file-mode 1))
	       (insert-file-contents xbm-file)))
	   (setq case-fold-search t)
	   (cond ((and
		   (search-forward "width" nil t)
		   (condition-case nil
		       (= 48 (read (current-buffer)))
		     (error nil))
		   (goto-char (point-min))
		   (search-forward "height" nil t)
		   (condition-case nil
		       (= 48 (read (current-buffer)))
		     (error nil))
		   (search-forward "{" nil t)
		   (progn
		     (delete-region (point-min) (point))
		     (while (search-forward "0x" nil t)
		       (replace-match ""))
		     (goto-char (point-min))
		     (while (re-search-forward "[^[:xdigit:]]+" nil t)
		       (replace-match ""))
		     (= 576 (buffer-size))))
		  (goto-char (point-min))
		  (insert "\"")
		  (while (not (eobp))
		    (insert "\\x")
		    (forward-char 2))
		  (insert "\"")
		  (goto-char (point-min))
		  (dolist (byte (prog1
				    (append (read (current-buffer)) nil)
				  (erase-buffer)))
		    (insert (format "%02x" (aref x-face-mirror byte))))
		  (call-process-region (point-min) (point-max)
				       compface-program t '(t nil))
		  (goto-char (point-min))
		  (while (search-forward "\r" nil t)
		    (replace-match ""))
		  (goto-char (point-min))
		  (skip-chars-forward "\t\n ")
		  (delete-region (point-min) (point))
		  (goto-char (point-max))
		  (skip-chars-backward "\t\n ")
		  (delete-region (point) (point-max)))
		 ((and (> (buffer-size) 0)
		       (progn
			 (goto-char (point-min))
			 (skip-chars-forward "\t\n\v\f\r -~")
			 (eobp)))
		  (goto-char (point-min))
		  (if (search-forward "\nX-Face" nil t)
		      ;; There may be multiple X-Faces.
		      (progn
			(goto-char (point-min))
			(skip-chars-forward "\t\n\r ")
			(delete-region (point-min) (point))
			(while (re-search-forward "[[:blank:]\r]+$" nil t)
			  (replace-match ""))
			(goto-char (point-max))
			(skip-chars-backward "\t\n ")
			(delete-region (point) (point-max)))
		    (while (re-search-forward "[[:blank:]\n\v\f\r]+" nil t)
		      (replace-match ""))
		    (goto-char (point-min))
		    (move-to-column 70)
		    (while (not (eobp))
		      (insert "\n ")
		      (move-to-column 78))))
		 (t
		  (x-face-insert-error 'file-error
				       "Unrecognized file format:" xbm-file)))
	   (buffer-string))))
    (set (make-local-variable 'x-face-use-overlay) t)
    (set (make-local-variable 'x-face-hide-related-headers) nil)
    (let ((props (default-value 'x-face-hidden-properties)))
      (set (make-local-variable 'x-face-hidden-properties)
	   (when (and x-face-field-icon
		      (plist-get props 'invisible))
	     (list 'display (copy-sequence x-face-field-icon)
		   'intangible (plist-get props 'intangible)))))
    (save-excursion
      (save-restriction
	(let ((inhibit-point-motion-hooks t))
	  (x-face-narrow-to-head)
	  (let (x-faces)
	    (let ((case-fold-search t))
	      (while (progn
		       (while (not (or (eobp)
				       (looking-at "X-Face:")))
			 (forward-line 1))
		       (not (eobp)))
		(push (cons (point)
			    (progn
			      (while (progn
				       (forward-line 1)
				       (memq (char-after) '(?\t ?\ ))))
			      (point)))
		      x-faces))
	      (goto-char (or (cdar x-faces) (point-max)))
	      (let ((inhibit-read-only t)
		    (keymap (current-local-map))
		    props start)
		(setq keymap (if keymap
				 (copy-keymap keymap)
			       (make-sparse-keymap)))
		(define-key keymap [mouse-2] 'x-face-show)
		(setq props (list 'local-map keymap
				  'mouse-face x-face-mouse-face
				  'help-echo
				  "mouse-2 toggles X-Face visibility"))
		(unless (bolp) (insert "\n"))
		(save-restriction
		  (narrow-to-region (point) (point))
		  (insert "X-Face: ")
		  (setq start (point))
		  (insert x-face "\n")
		  (goto-char start)
		  (forward-line 1)
		  (while (not (eobp))
		    (while (memq (char-after) '(?\t ?\ ))
		      (forward-line 1))
		    (x-face-set-properties start (1- (point)) props)
		    (while (prog1
			       (not (or (eobp)
					(and (looking-at
					      "X-Face:[[:blank:]\n]+")
					     (setq start (match-end 0)))))
			     (forward-line 1)))))
		(unless keep-existing-headers
		  (while (setq x-face (pop x-faces))
		    (delete-region (car x-face) (cdr x-face)))))))))))
  (when (if (functionp x-face-auto-image)
	    (funcall x-face-auto-image)
	  x-face-auto-image)
    (x-face-show 1)))

;;;###autoload
(defun x-face-save (&optional filename)
  "Save X-Face headers to XBM files.  This requires the external program
`uncompface'.  It currently does not work with the forwarded messages
except for Mew."
  (interactive)
  (let ((case-fold-search t)
	x-faces gray from)
    (save-excursion
      (x-face-possibly-change-buffer-for-saving)
      (save-restriction
	(x-face-narrow-to-head)
	(let ((inhibit-point-motion-hooks t)
	      start)
	  (while (progn
		   (while (not (or (eobp)
				   (looking-at "X-Face\\(-[[:digit:]]+\\)?:")))
		     (forward-line 1))
		   (not (eobp)))
	    (setq start (point))
	    (when (match-beginning 1)
	      ;; There are gray X-Faces.
	      (setq gray t))
	    (while (progn
		     (forward-line 1)
		     (memq (char-after) '(?\t ?\ ))))
	    (push (buffer-substring-no-properties start (point)) x-faces))
	  (when (setq from (mail-fetch-field "from"))
	    (setq from (cadr (mail-extract-address-components from)))))))
    (when (x-face-buffer-live-p x-face-working-buffer)
      (kill-buffer x-face-working-buffer))
    (if x-faces
	(let* ((minibuffer-setup-hook (append minibuffer-setup-hook
					      '(end-of-line)))
	       (x-face-read-file-name-type (if gray
					       'xpm
					     'xbm))
	       (default (format "%s.%s%s"
				(or from "x-face")
				x-face-read-file-name-type
				(or (cdr (assq x-face-compressor
					       '((bzip2 . ".bz2")
						 (gzip . ".gz"))))
				    "")))
	       (num 0)
	       (log-buffer (get-buffer-create "*X-Face-Save-Log*"))
	       directory basename basename-re nform suffix compr)
	  (unless filename
	    (setq filename
		  (x-face-read-file-name
		   (concat (upcase (format "%s" x-face-read-file-name-type))
			   " file: ")
		   x-face-image-file-directory-for-save
		   default nil default)))
	  (setq directory (or (file-name-directory filename)
			      x-face-image-file-directory-for-save)
		filename (file-name-nondirectory filename)
		basename (if (string-match
			      (format "\
\\(?:-[[:digit:]]+\\)?\\.%s\\(?:\\.bz2\\|\\.gz\\)?$"
				      x-face-read-file-name-type)
			      filename)
			     (substring filename 0 (match-beginning 0))
			   filename)
		basename-re (concat "^" (regexp-quote basename)))
	  (if (file-directory-p directory)
	      (let ((files (directory-files directory nil basename-re t)))
		(when files
		  (let ((re (concat basename-re "-\\([[:digit:]]+\\)"))
			file)
		    (while (setq file (pop files))
		      (setq num (max num (if (string-match re file)
					     (1+ (string-to-number
						  (match-string 1 file)))
					   1)))))))
	    (make-directory directory t))
	  (with-current-buffer log-buffer
	    (setq case-fold-search nil)
	    (goto-char (point-max))
	    (unless (bolp) (insert "\n"))
	    (buffer-disable-undo))
	  (setq nform (concat "%s-%0" (number-to-string
				       (max 2 (length (number-to-string num))))
			      "d"))
	  (when (string-match (format "\\(\\.%s\\)\\(\\.bz2\\|\\.gz\\)?$"
				      x-face-read-file-name-type)
			      filename)
	    (setq suffix (match-string 1 filename)
		  compr (match-string 2 filename)))
	  (while x-faces
	    (setq filename (if (zerop num)
			       basename
			     (format nform basename num))
		  num (1+ num))
	    (with-temp-buffer
	      (if gray
		  (progn
		    (insert (x-face-gray-x-faces-to-xpm x-faces nil filename))
		    (setq x-faces nil))
		(insert (x-face-to-xbm (pop x-faces) filename)))
	      (setq filename (expand-file-name (concat filename suffix compr)
					       directory))
	      (if (and compr
		       (require 'jka-compr)
		       (not auto-compression-mode))
		  (progn
		    (auto-compression-mode 1)
		    (unwind-protect
			(write-region (point-min) (point-max) filename)
		      (auto-compression-mode -1)))
		(write-region (point-min) (point-max) filename))
	      (when (interactive-p)
		(x-face-put-file-name-in-history filename)
		(when x-faces
		  (sit-for 0.5)))
	      (with-current-buffer log-buffer
		(insert (format-time-string "%Y-%m-%d %T ")
			(abbreviate-file-name filename)
			"\n")
		(bury-buffer log-buffer)))))
      (message "No X-Face headers found"))))

(eval-and-compile
  (autoload 'bbdb-current-record "bbdb-com")
  (autoload 'bbdb-record-getprop "bbdb")
  (autoload 'bbdb-record-name "bbdb"))

(defun x-face-energize-bbdb-buffer (&optional all silent)
  "Show X-Face images in the BBDB buffer.  If optional ALL is not
specified, it deals with only the current record."
  (when x-face-bbdb-display
    (save-excursion
      (set-buffer (symbol-value 'bbdb-buffer-name))
      (save-restriction
	(unless all
	  (beginning-of-line)
	  (narrow-to-region (progn
			      (while (and (memq (char-after)
						'(nil ?\t ?\n ?\ ))
					  (zerop (forward-line -1))))
			      (point))
			    (progn
			      (while (progn
				       (forward-line 1)
				       (memq (char-after) '(?\t ?\ ))))
			      (point))))
	(x-face-remove-x-face-images)
	(when (and x-face-bbdb-display
		   (display-images-p)
		   (image-type-available-p 'pbm))
	  (let ((buffer (get-buffer-create " *x-face-energize-bbdb*"))
		(x-face-use-overlay t)
		(picon (when x-face-field-icon
			 (let* ((icon (copy-sequence x-face-field-icon))
				(type (plist-get (cdr icon) :type))
				(bgmode (if (x-face-dark-background-p)
					    'dark
					  'light))
				(params
				 (x-face-cleanup-plist
				  (cdr (assq bgmode x-face-image-attributes))))
				(bg (plist-get params :background))
				(fg (plist-get params :foreground)))
			   (unless (stringp bg)
			     (unless (stringp
				      (setq bg (face-background 'default)))
			       (setq bg "#ffffff")))
			   (unless (stringp fg)
			     (unless (stringp
				      (setq fg (face-foreground 'default)))
			       (setq fg "#000000")))
			   (cond ((eq 'xpm type)
				  (plist-put (cdr icon) :color-symbols
					     (list (cons "outline"
							 (if (eq 'dark bgmode)
							     bg
							   fg)))))
				 ((memq type '(pbm xbm))
				  (plist-put (cdr icon) :foreground
					     (if (eq 'dark bgmode)
						 bg
					       fg))))
			   icon)))
		(x-face-image-attributes (x-face-image-attributes-for-bbdb))
		;; Space with the face `default' used to stop leaking
		;; of neighboring faces.
		(space (propertize " " 'face 'default))
		record faces msg nfaces start end from to icon)
	    (goto-char (point-min))
	    (unwind-protect
		(while (not (eobp))
		  (when all
		    ;; Go to the beginning or the next recored.
		    (while (and (looking-at "[[:blank:]]\\|[[:blank:]]*$")
				(zerop (forward-line 1)))))
		  (when (and (not (eobp))
			     (setq record (bbdb-current-record))
			     (setq faces (bbdb-record-getprop record 'face)))
		    (when (and all
			       (not silent))
		      (setq msg (format "Extracting X-Face(s) for %s..."
					(bbdb-record-name record)))
		      (message "%s" msg))
		    (setq faces (split-string faces)
			  nfaces (length faces))
		    (with-current-buffer buffer
		      (erase-buffer)
		      (insert "From: X\n")
		      (dolist (face faces)
			(insert "X-Face: " face "\n"))
		      (x-face-decode-message-header)
		      (setq faces (overlay-properties (car (overlays-at 6)))))
		    (plist-put faces 'before-string
			       (concat (plist-get faces 'before-string)
				       space))
		    (x-face-set-properties (point) (1+ (point)) faces)
		    (forward-line 1)
		    (when picon
		      ;; Iconify the raw X-Face records.
		      (while (not (or (when (looking-at "[[:blank:]]+face:")
					(goto-char (match-end 0)))
				      (looking-at "[^[:blank:]\n]*$")
				      (eobp)))
			(forward-line 1))
		      (unless (or (bolp) (eobp))
			(setq start (point))
			(while (progn
				 (forward-line 1)
				 ;; There might not be a blank line between
				 ;; records when BBDB v2.33 and later is
				 ;; running.
				 (not (or (looking-at "\
\[^[:blank:]\n]\\|[[:blank:]]+[^[:blank:]\n:]+:\
\\(?:[[:blank:]]\\|[[:blank:]]*$\\)")
					  (eobp)))))
			(setq end (point))
			(goto-char start)
			(while (re-search-forward "[^[:blank:]\n]+" end t)
			  (setq from (match-beginning 0)
				to (match-end 0)
				nfaces (1- nfaces))
			  (if (zerop nfaces)
			      ;; Use the parent icon directly.
			      (setq icon picon)
			    (when (and (looking-at "[[:blank:]\n]+")
				       (< (match-end 0) end))
			      (setq to (match-end 0)
				    icon (copy-sequence picon))))
			  (x-face-set-properties from to
						 (list 'display icon
						       'intangible t
						       'x-face-image t))))))
		  (forward-line 1))
	      (when msg
		(message "%sdone" msg))
	      (kill-buffer buffer))))))))

(let (current-load-list)
  (defadvice bbdb-display-records-1
    (around show-x-face-images-in-the-bbdb-buffer activate compile)
    "Advised by X-Face-E21.  Show X-Faces images in the BBDB buffer."
    (let ((silent (or (and (boundp 'bbdb-gag-messages)
			   (symbol-value 'bbdb-gag-messages))
		      (and (boundp 'bbdb-silent-running)
			   (symbol-value 'bbdb-silent-running)))))
      (unless silent
	(message "Formatting..."))
      (let ((bbdb-silent-running t)
	    (bbdb-list-hook bbdb-list-hook))
	(remove-hook 'bbdb-list-hook 'x-face-energize-bbdb-buffer)
	ad-do-it)
      (x-face-energize-bbdb-buffer 'all silent)
      (unless silent
	(message "Formatting...done")))))

(add-hook 'bbdb-list-hook 'x-face-energize-bbdb-buffer)

(provide 'x-face-e21)

(eval-when-compile
  (when noninteractive
    (message "%s" "\r
  ++ Do `isearch' for the string \"STRONGLY\" in x-face-e21.el\r
  ++ and see around there for your happy networking.\r
  ++ (It is not an error message, never mind :-)\r
")))

;;; x-face-e21.el ends here
